@@ -20,6 +20,8 @@
  *
  */
 
-void resolve_add_dns(uint32_t ifindex, uint8_t type, char **dns_list);
-void resolve_add_domain_name(uint32_t ifindex, const char *domain_name);
-void resolve_remove(uint32_t ifindex);
+struct resolve *resolve_new(uint32_t ifindex);
+void resolve_set_dns(struct resolve *resolve, char **dns_list);
+void resolve_set_domains(struct resolve *resolve, char **domain_list);
+void resolve_revert(struct resolve *resolve);
+void resolve_free(struct resolve *resolve);
