@@ -247,9 +247,6 @@ static void wsc_enrollee_netdev_event(struct netdev *netdev,
 	case NETDEV_EVENT_AUTHENTICATING:
 	case NETDEV_EVENT_ASSOCIATING:
 		break;
-	case NETDEV_EVENT_LOST_BEACON:
-		l_debug("Lost beacon");
-		break;
 	case NETDEV_EVENT_DISCONNECT_BY_AP:
 		l_debug("Disconnect by AP");
 		wsc_enrollee_connect_cb(wsce->netdev,
@@ -262,7 +259,7 @@ static void wsc_enrollee_netdev_event(struct netdev *netdev,
 	default:
 		l_debug("Unexpected event: %d", event);
 		break;
-	};
+	}
 }
 
 static void wsc_enrollee_handshake_event(struct handshake_state *hs,
@@ -536,15 +533,10 @@ static void wsc_store_credentials(struct wsc_credentials_info *creds,
 		l_debug("Storing credential for '%s(%s)'", ssid,
 						security_to_str(security));
 
-		if (security == SECURITY_PSK) {
-			char *hex = l_util_hexstring(creds[i].psk,
-						sizeof(creds[i].psk));
-
-			l_settings_set_value(settings, "Security",
-							"PreSharedKey", hex);
-			explicit_bzero(hex, strlen(hex));
-			l_free(hex);
-		}
+		if (security == SECURITY_PSK)
+			l_settings_set_bytes(settings, "Security",
+						"PreSharedKey", creds[i].psk,
+                                                sizeof(creds[i].psk));
 
 		storage_network_sync(security, ssid, settings);
 		l_settings_free(settings);
@@ -1111,7 +1103,7 @@ static struct l_dbus_message *wsc_push_button(struct l_dbus *dbus,
 	if (!l_dbus_message_get_arguments(message, ""))
 		return dbus_error_invalid_args(message);
 
-	if (wsc->pending_connect)
+	if (wsc->pending_connect || wsc->pending_cancel)
 		return dbus_error_busy(message);
 
 	wsc->pending_connect = l_dbus_message_ref(message);
@@ -1137,6 +1129,7 @@ static struct l_dbus_message *wsc_generate_pin(struct l_dbus *dbus,
 
 	reply = l_dbus_message_new_method_return(message);
 	l_dbus_message_set_arguments(reply, "s", pin);
+	explicit_bzero(pin, 9);
 
 	return reply;
 }
@@ -1150,7 +1143,7 @@ static struct l_dbus_message *wsc_start_pin(struct l_dbus *dbus,
 
 	l_debug("");
 
-	if (wsc->pending_connect)
+	if (wsc->pending_connect || wsc->pending_cancel)
 		return dbus_error_busy(message);
 
 	if (!l_dbus_message_get_arguments(message, "s", &pin))
@@ -1301,3 +1294,4 @@ static void wsc_exit(void)
 }
 
 IWD_MODULE(wsc, wsc_init, wsc_exit)
+IWD_MODULE_DEPENDS(wsc, netdev);
